@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2009-2011 Nick Mathewson and Niels Provos
+ * Copyright (c) 2009-2012 Nick Mathewson and Niels Provos
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -469,7 +469,7 @@ check_error_logging(void (*fn)(void), int wantexitcode,
 	int status = 0, exitcode;
 	fatal_want_severity = wantseverity;
 	fatal_want_message = wantmsg;
-	if ((pid = fork()) == 0) {
+	if ((pid = regress_fork()) == 0) {
 		/* child process */
 		fn();
 		exit(0); /* should be unreachable. */
